# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_celery_fulldbresult', '0002_auto_20150713_1520'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskresultmeta',
            name='eta',
            field=models.DateTimeField(verbose_name='eta', null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import djcelery.picklefield


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TaskResultMeta',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, verbose_name='ID', serialize=False)),
                ('task_id', models.CharField(unique=True, verbose_name='task id', max_length=255)),
                ('task', models.CharField(verbose_name='task name', max_length=200)),
                ('args', models.TextField(help_text='JSON encoded positional arguments', verbose_name='Arguments', default='[]', blank=True)),
                ('kwargs', models.TextField(help_text='JSON encoded keyword arguments', verbose_name='Keyword arguments', default='{}', blank=True)),
                ('hostname', models.CharField(verbose_name='hostname', max_length=255, blank=True, null=True)),
                ('exchange', models.CharField(verbose_name='exchange', max_length=200, default=None, blank=True, null=True)),
                ('routing_key', models.CharField(verbose_name='routing key', max_length=200, default=None, blank=True, null=True)),
                ('expires', models.DateTimeField(verbose_name='expires', blank=True, null=True)),
                ('status', models.CharField(verbose_name='state', max_length=50, default='PENDING', choices=[('REVOKED', 'REVOKED'), ('SUCCESS', 'SUCCESS'), ('RECEIVED', 'RECEIVED'), ('PENDING', 'PENDING'), ('RETRY', 'RETRY'), ('FAILURE', 'FAILURE'), ('STARTED', 'STARTED')])),
                ('result', djcelery.picklefield.PickledObjectField(default=None, editable=False, null=True)),
                ('date_submitted', models.DateTimeField(verbose_name='submitted at', blank=True, null=True)),
                ('date_done', models.DateTimeField(verbose_name='done at', auto_now=True)),
                ('traceback', models.TextField(verbose_name='traceback', blank=True, null=True)),
                ('hidden', models.BooleanField(db_index=True, default=False, editable=False)),
                ('meta', djcelery.picklefield.PickledObjectField(default=None, editable=False, null=True)),
            ],
            options={
                'verbose_name_plural': 'task states',
                'verbose_name': 'task state',
                'db_table': 'celery_taskresultmeta',
            },
            bases=(models.Model,),
        ),
    ]



from __future__ import absolute_import, unicode_literals

from celery.states import SUCCESS, FAILURE
from djcelery.managers import TaskManager, transaction_retry

from django.utils.timezone import now


DEFAULT_ARGS = "[]"
DEFAULT_KWARGS = "{}"
TERMINAL_STATES = (
    SUCCESS,
    FAILURE
)


class TaskResultManager(TaskManager):

    def get_stale_scheduled_tasks(self, max_duration=None):
        """Selects scheduled tasks, which are stale: they have an old ETA, they
        have a scheduled state and a scheduled id, but probably were never sent
        for execution.
        """
        from django_celery_fulldbresult.models import SCHEDULED

        max_date = now()
        if max_duration:
            max_date = now() - max_duration

        return self.filter(
            eta__lte=max_date, status=SCHEDULED,
            scheduled_id__isnull=False)

    def get_stale_tasks(self, max_duration=None, acceptable_states=None):
        """Selects tasks, which are stale: they were last updated before
        now-max_duration and they are not in one of the acceptable_states.

        :keyword max_duration: A timedelta. If None, the timedelta is 0.
        :keyword acceptable_states: A collection of states that are considered
            terminal. Tasks that are not in an acceptable state are
            considered stale. Default to (SUCCESS, FAILURE)
        :rtype: QuerySet
        """
        max_date = now()
        if max_duration:
            max_date = now() - max_duration

        if not acceptable_states:
            acceptable_states = TERMINAL_STATES

        return self.filter(date_done__lte=max_date).exclude(
            status__in=acceptable_states)

    @transaction_retry(max_retries=2)
    def store_result(
            self, task_id, result, status, traceback=None, children=None,
            task=None, args=DEFAULT_ARGS, kwargs=DEFAULT_KWARGS, expires=None,
            routing_key=None, exchange=None, hostname=None,
            date_submitted=None, eta=None):
        """Store the result and status of a task.
        :param task_id: task id
        :param result: The return value of the task, or an exception
            instance raised by the task.
        :param status: Task status. See
            :meth:`celery.result.AsyncResult.get_status` for a list of
            possible status values.
        :keyword traceback: The traceback at the point of exception (if the
            task failed).
        :keyword children: List of serialized results of subtasks
            of this task.
        :keyword exception_retry_count: How many times to retry by
            transaction rollback on exception. This could theoretically
            happen in a race condition if another worker is trying to
            create the same task. The default is to retry twice.
        :keyword task: todo
        :keyword args: todo
        :keyword kwargs: todo
        :keyword expires: todo
        :keyword routing_key: todo
        :keyword exchange: todo
        :keyword hostname: todo
        :keyword date_submitted: todo
        :keyword eta: todo
        """
        defaults = {
            "status": status,
            "result": result,
            "traceback": traceback,
            "task": task,
            "args": args,
            "kwargs": kwargs,
            "expires": expires,
            "routing_key": routing_key,
            "exchange": exchange,
            "hostname": hostname,
            "eta": eta,
            "meta": {
                "children": children
            }
        }

        if date_submitted is not None:
            # Only set date_submitted if not None
            defaults["date_submitted"] = date_submitted

        return self.update_or_create(
            task_id=task_id,
            defaults=defaults
        )



from __future__ import absolute_import, unicode_literals

import json

from djcelery.backends.database import DatabaseBackend

from django_celery_fulldbresult.models import TaskResultMeta


class DatabaseResultBackend(DatabaseBackend):
    """Database backend that stores enough task metadata to retry the task.
    """

    TaskModel = TaskResultMeta

    def _store_result(self, task_id, result, status, traceback, request=None):
        if request:
            # TODO Use celery json serializer
            args = json.dumps(request.args)
            kwargs = json.dumps(request.kwargs)
            # Sometimes, request.task is an object and not a str, so the right
            # solution is:
            # 1. if request.name exists (this is a Request object), use it.
            # 2. Otherwise, use request.task (this will be a string from the
            # Context object)
            task = getattr(request, "name", None) or request.task
            expires = request.expires
            delivery_info = request.delivery_info or {}
            routing_key = delivery_info.get("routing_key")
            exchange = delivery_info.get("exchange")
            hostname = request.hostname
            date_submitted = getattr(request, "date_submitted", None)
            eta = request.eta
        else:
            args = []
            kwargs = {}
            task = ""
            expires = None
            routing_key = None
            exchange = None
            hostname = None
            date_submitted = None
            eta = None
        self.TaskModel._default_manager.store_result(
            task_id, result, status,
            traceback=traceback, children=self.current_task_children(request),
            task=task, args=args, kwargs=kwargs, expires=expires,
            routing_key=routing_key, exchange=exchange, hostname=hostname,
            date_submitted=date_submitted, eta=eta)
        return result



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django_celery_fulldbresult.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_celery_fulldbresult', '0003_taskresultmeta_eta'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskresultmeta',
            name='result',
            field=django_celery_fulldbresult.models.PickledOrJSONObjectField(default=None, null=True, editable=False),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.8 on 2016-07-22 19:53
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_celery_fulldbresult', '0004_auto_20150914_1947'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskresultmeta',
            name='scheduled_id',
            field=models.UUIDField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='taskresultmeta',
            name='status',
            field=models.CharField(choices=[('FAILURE', 'FAILURE'), ('PENDING', 'PENDING'), ('RECEIVED', 'RECEIVED'), ('RETRY', 'RETRY'), ('REVOKED', 'REVOKED'), ('SCHEDULED', 'SCHEDULED'), ('SCHEDULED_SENT', 'SCHEDULED_SENT'), ('STARTED', 'STARTED'), ('SUCCESS', 'SUCCESS')], default='PENDING', max_length=50, verbose_name='state'),
        ),
    ]



from __future__ import absolute_import, unicode_literals


class SchedulingStopPublishing(Exception):
    """Raised before publishing a scheduled task to prevent Celery from sending
    the task to the broker.
    """
    def __init__(self, task_id):
        super(SchedulingStopPublishing, self).__init__()
        self.task_id = task_id



from __future__ import absolute_import, unicode_literals

from celery import current_app
from celery.states import PENDING
from celery.app.task import Context, Task
from celery.signals import before_task_publish
from django_celery_fulldbresult.errors import SchedulingStopPublishing

from django.conf import settings
from django.utils.timezone import now


old_apply_async = Task.apply_async


def new_apply_async(self, *args, **kwargs):
    try:
        return old_apply_async(self, *args, **kwargs)
    except SchedulingStopPublishing as exc:
        # There was an ETA and the task was not sent to the broker.
        # A scheduled task was created instead.
        return self.AsyncResult(exc.task_id)


Task.apply_async = new_apply_async

schedule_eta = getattr(
    settings, "DJANGO_CELERY_FULLDBRESULT_SCHEDULE_ETA", False)

track_publish = getattr(
    settings, "DJANGO_CELERY_FULLDBRESULT_TRACK_PUBLISH", False)

if track_publish or schedule_eta:
    @before_task_publish.connect
    def update_sent_state(sender=None, body=None, exchange=None,
                          routing_key=None, **kwargs):
        # App may not be loaded on init
        from django_celery_fulldbresult.models import SCHEDULED

        task = current_app.tasks.get(sender)
        save = False
        status = None

        schedule_eta = getattr(
            settings, "DJANGO_CELERY_FULLDBRESULT_SCHEDULE_ETA", False)

        track_publish = getattr(
            settings, "DJANGO_CELERY_FULLDBRESULT_TRACK_PUBLISH", False)

        ignore_result = getattr(task, "ignore_result", False) or\
            getattr(settings, "CELERY_IGNORE_RESULT", False)

        if schedule_eta and body.get("eta") and not body.get("chord")\
                and not body.get("taskset"):
            status = SCHEDULED
            save = True
        elif track_publish and not ignore_result:
            status = PENDING
            save = True

        if save:
            backend = task.backend if task else current_app.backend
            request = Context()
            request.update(**body)
            request.date_submitted = now()
            request.delivery_info = {
                "exchange": exchange,
                "routing_key": routing_key
            }
            backend.store_result(
                body["id"], None, status, traceback=None, request=request)

        if status == SCHEDULED:
            raise SchedulingStopPublishing(task_id=body["id"])



from __future__ import absolute_import, unicode_literals

import json

from django.conf import settings
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from celery import current_app
from djcelery.admin import PeriodicTaskAdmin
from djcelery.models import PeriodicTask

from django_celery_fulldbresult.models import TaskResultMeta


def retry_task(modeladmin, request, queryset):
    """Admin Site Action that retries the selected tasks.
    """
    task_ids = []
    for task in queryset.all():
        task_name = task.task
        task_args = json.loads(task.args)
        task_kwargs = json.loads(task.kwargs)
        result = current_app.send_task(
            task_name, args=task_args, kwargs=task_kwargs)
        task_ids.append(result.task_id)
    modeladmin.message_user(request, "Tasks sent: {0}".format(task_ids))
retry_task.short_description = _("Retry Tasks")


class TaskResultMetaAdmin(admin.ModelAdmin):
    list_display = ("task_id", "task", "date_submitted", "date_done", "status",
                    "eta")
    search_fields = ("task_id", "task")
    list_filter = ("status", "task")
    actions = [retry_task, ]
    readonly_fields = ("result_repr", )


admin.site.register(TaskResultMeta, TaskResultMetaAdmin)


# Option to override PeriodicTaskAdmin
def trigger_periodic_task(modeladmin, request, queryset):
    """Admin Site Action for django_celery's PeriodicTask that manually
    triggers the selected tasks.
    """
    retry_task(modeladmin, request, queryset)
trigger_periodic_task.short_description = _("Trigger Periodic Tasks")


class CustomPeriodicTaskAdmin(PeriodicTaskAdmin):
    actions = [trigger_periodic_task, ]
    list_filter = ("task", )


if getattr(settings, "DJANGO_CELERY_FULLDBRESULT_OVERRIDE_DJCELERY_ADMIN",
           False):
    admin.site.unregister(PeriodicTask)
    admin.site.register(PeriodicTask, CustomPeriodicTaskAdmin)



import json
from uuid import uuid4

from celery import shared_task, current_app

from django.utils.timezone import now
from django_celery_fulldbresult.models import (
    TaskResultMeta, SCHEDULED, SCHEDULED_SENT)


@shared_task(ignore_result=True)
def send_scheduled_task():
    """Task that sends due scheduled tasks for execution.

    Each DB operation must be in its own commit so that even if
    send_scheduled_task is called multiple times concurrently, a task is
    ensured to only be sent **at most once**. If a crash occurs while sending a
    task, the task will stay indefinitely in the SCHEDULED state while having a
    schedule id.

    1. Tasks due to be executed are marked with a schedule id. This prevents
        the task from being sent for execution twice.

    2. Each task marked by the schedule id is sent for exeuction without an
        ETA.

    3. We change the status from SCHEDULED to SCHEDULED SENT after a task is
        being sent for execution.

    **IMPORTANT: ** Never call this task inside an atomic block or you could
    end up sending tasks more than once. Always use autocommit (each SQL query
    executed in its own transaction).
    """
    limit = now()
    schedule_id = uuid4().hex

    # Mark tasks ready to be scheduled
    TaskResultMeta.objects.filter(
        scheduled_id__isnull=True, eta__lt=limit,
        status=SCHEDULED).update(
        scheduled_id=schedule_id)

    # Fetch and apply by removing eta
    for task in TaskResultMeta.objects.filter(
            scheduled_id=schedule_id).all():
        task_name = task.task
        task_args = json.loads(task.args)
        task_kwargs = json.loads(task.kwargs)
        result = current_app.send_task(
            task_name, args=task_args, kwargs=task_kwargs)
        task.status = SCHEDULED_SENT
        task.result = {"new_task_id": result.task_id}
        task.save()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_celery_fulldbresult', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskresultmeta',
            name='args',
            field=models.TextField(help_text='JSON encoded positional arguments', blank=True, verbose_name='arguments', default='[]'),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='taskresultmeta',
            name='kwargs',
            field=models.TextField(help_text='JSON encoded keyword arguments', blank=True, verbose_name='keyword arguments', default='{}'),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='taskresultmeta',
            name='status',
            field=models.CharField(choices=[('FAILURE', 'FAILURE'), ('PENDING', 'PENDING'), ('RECEIVED', 'RECEIVED'), ('RETRY', 'RETRY'), ('REVOKED', 'REVOKED'), ('STARTED', 'STARTED'), ('SUCCESS', 'SUCCESS')], verbose_name='state', default='PENDING', max_length=50),
            preserve_default=True,
        ),
    ]



from __future__ import absolute_import, unicode_literals

import json

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_unicode as force_text  # noqa


from celery import states

from djcelery.compat import python_2_unicode_compatible
from djcelery.picklefield import PickledObjectField


from django_celery_fulldbresult import managers

SCHEDULED = "SCHEDULED"

SCHEDULED_SENT = "SCHEDULED_SENT"

FULL_RESULTS_ALL_STATES = states.ALL_STATES.union(
    frozenset([SCHEDULED, SCHEDULED_SENT]))


TASK_STATE_CHOICES = sorted(
    zip(FULL_RESULTS_ALL_STATES, FULL_RESULTS_ALL_STATES), key=lambda t: t[0])


def use_json():
    """Returns True if django celery db result is configured to save results as
    JSON.
    """
    return getattr(settings, "DJANGO_CELERY_FULLDBRESULT_USE_JSON", False)


def force_json():
    """Returns True if django celery db result is configured to force JSON even
    if a result is not JSON serializable.
    """
    return getattr(settings, "DJANGO_CELERY_FULLDBRESULT_FORCE_JSON", False)


class PickledOrJSONObjectField(PickledObjectField):
    """Serializes field content using pickle or JSON depending on the
    DJANGO_CELERY_FULLDBRESULT_USE_JSON.
    """

    def get_db_prep_value(self, value, **kwargs):
        if use_json():
            if value is not None:
                try:
                    value = force_text(json.dumps(value))
                except TypeError:
                    if force_json():
                        value = force_text(
                            json.dumps(
                                {
                                    "value": str(value),
                                    "forced_json": True
                                }
                            )
                        )
                    else:
                        raise
        else:
            value = super(PickledOrJSONObjectField, self).get_db_prep_value(
                value, **kwargs)

        return value

    def to_python(self, value):
        if use_json():
            if value:
                try:
                    value = json.loads(value)
                    return value
                except Exception:
                    if isinstance(value, str):
                        # Badly formatted JSON!
                        raise
                    else:
                        return value
            else:
                return None
        else:
            return super(PickledOrJSONObjectField, self).to_python(value)


@python_2_unicode_compatible
class TaskResultMeta(models.Model):
    """Task result/status.
    """

    task_id = models.CharField(_("task id"), max_length=255, unique=True)
    task = models.CharField(_("task name"), max_length=200)
    args = models.TextField(
        _("arguments"), blank=True, default="[]",
        help_text=_("JSON encoded positional arguments"),
    )
    kwargs = models.TextField(
        _("keyword arguments"), blank=True, default="{}",
        help_text=_("JSON encoded keyword arguments"),
    )
    hostname = models.CharField(_('hostname'), max_length=255, blank=True,
                                null=True)
    exchange = models.CharField(
        _('exchange'), max_length=200, blank=True, null=True, default=None,
    )
    routing_key = models.CharField(
        _('routing key'), max_length=200, blank=True, null=True, default=None,
    )
    eta = models.DateTimeField(
        _('eta'), blank=True, null=True,
    )
    expires = models.DateTimeField(
        _('expires'), blank=True, null=True,
    )
    status = models.CharField(
        _("state"),
        max_length=50, default=states.PENDING, choices=TASK_STATE_CHOICES,
    )
    scheduled_id = models.UUIDField(null=True, blank=True)
    result = PickledOrJSONObjectField(null=True, default=None, editable=False)
    date_submitted = models.DateTimeField(
        _("submitted at"), null=True, blank=True)
    date_done = models.DateTimeField(_("done at"), auto_now=True)
    traceback = models.TextField(_("traceback"), blank=True, null=True)
    hidden = models.BooleanField(editable=False, default=False, db_index=True)
    # TODO compression was enabled by mistake in Celery,
    # djcelery recommends to disable it
    # but this is a backwards incompatible change that needs planning.
    meta = PickledObjectField(
        compress=True, null=True, default=None, editable=False,
    )

    objects = managers.TaskResultManager()

    @property
    def result_repr(self):
        return str(self.result)

    class Meta:
        verbose_name = _("task state")
        verbose_name_plural = _("task states")
        db_table = "celery_taskresultmeta"

    def to_dict(self):
        return {"task_id": self.task_id,
                "status": self.status,
                "result": self.result,
                "date_done": self.date_done,
                "traceback": self.traceback,
                "children": (self.meta or {}).get("children")}

    def __str__(self):
        return "<Task: {0.task_id} state={0.status}>".format(self)



from datetime import timedelta

from django.core.management import BaseCommand

from django_celery_fulldbresult.models import TaskResultMeta


class Command(BaseCommand):
    option_list = BaseCommand.option_list

    def add_arguments(self, parser):
        parser.add_argument(
            "--days",
            action="store",
            dest="days",
            type=int,
            default=0,
            help="max days before a scheduled task is stale")
        parser.add_argument(
            "--seconds",
            action="store",
            dest="seconds",
            type=int,
            default=0,
            help="max seconds before a task is stale")
        parser.add_argument(
            "--microseconds",
            action="store",
            dest="microseconds",
            type=int,
            default=0,
            help="max microseconds before a task is stale")
        parser.add_argument(
            "--minutes",
            action="store",
            dest="minutes",
            type=int,
            default=0,
            help="max minutes before a task is stale")
        parser.add_argument(
            "--hours",
            action="store",
            dest="hours",
            type=int,
            default=0,
            help="max hours before a task is stale")
        parser.add_argument(
            "--weeks",
            action="store",
            dest="weeks",
            type=int,
            default=0,
            help="max weeks before a task is stale")

    def handle(self, *args, **options):
        delta = timedelta(
            days=options["days"], seconds=options["seconds"],
            microseconds=options["microseconds"], minutes=options["minutes"],
            hours=options["hours"], weeks=options["weeks"])

        tasks = TaskResultMeta.objects.get_stale_scheduled_tasks(
            delta)
        print("Stale scheduled tasks:")
        for task in tasks:
            print("  {0} - {1}: {2}".format(
                task.date_done, task.task_id, task.task))
